
#include "messagedecoder.h"




message_decoder::message_decoder(decoder_application_configuration application_parameters, frame_specification frame_parameters)
{
	screen_width=frame_parameters.screen_width;
	screen_height=frame_parameters.screen_height;
	raster=frame_parameters.raster;
	next_pixel=0;
	total_nof_pixels=screen_width*screen_height;
	
	encoded_gif=frame_parameters.encoded_gif;
	next_bit=0;
	encoded_gif_length=frame_parameters.encoded_gif_length;
	
	BPP=frame_parameters.BPP;
	bits_interval=application_parameters.bits_interval;
	bits_length_reduction=application_parameters.bits_length_reduction;
	output_message_file=application_parameters.output_msg_file;
	
	cc = 1 << BPP;
	eoi=cc+1;
	freecode=eoi+1;
	CodeSize=BPP+1;

	axon=new unsigned short[5000];
	next=new unsigned short[5000];
	pix=new unsigned char[5000];
	iniroots();
	
	number_of_bits_of_message=0;
	nofphrases=0;
	nofencodedphrases=0;
	phrase_counter=0;
	
	msg=new unsigned char[1024*1024];
	for (int i=0;i<1024*1024;i++)
	{
		msg[i]=0;
	}
	
	Prefix = new unsigned short[0x2000];

	FirstEntry=eoi+1;
	NextEntry=FirstEntry;
	next_pixel=0;

}

message_decoder::~message_decoder()
{
	delete[] msg;
	delete[] axon;
	delete[] next;
	delete[] pix;
	delete[] Prefix;

}

void message_decoder::iniroots()
{
	unsigned short nofrootcodes=1<<max(2,BPP);
	for(unsigned short i=0; i<nofrootcodes; i++)
	{
		axon[i]=0;
		pix[i]=(unsigned char)i;
		// next[] is unused for root codes       
	}
	freecode=eoi+1;
} // gifcompressor::iniroots


unsigned short message_decoder::next_phrase_nongreedy_from_file()
{	
	long LongCode;
	LongCode= *((long*) (encoded_gif + (next_bit / 8)));
	LongCode= (LongCode >> (next_bit % 8));
	unsigned short Code;
	Code=LongCode & ((1 << CodeSize)-1);
	next_bit=next_bit+CodeSize;
	return Code;
};

int message_decoder::length_of_next_phrase_greedy()
{
	int temp_next_pixel=next_pixel;
	unsigned char pixel=raster[temp_next_pixel];
	temp_next_pixel++;
	unsigned short up=pixel;
	unsigned short down;
	int length=1;
	pixel=raster[temp_next_pixel];
	temp_next_pixel++;
	while ((down=findpixeloutlet(up,pixel)) != 0 )
	{
		up=down;
		if (temp_next_pixel >= total_nof_pixels) break;
		pixel=raster[temp_next_pixel];
		temp_next_pixel++;
		length++;
	}
	return length;
};

unsigned short message_decoder::findpixeloutlet(unsigned short headnode,unsigned char pixel)
{
	unsigned short outlet=axon[headnode];
	while(outlet)
	{
		if (pix[outlet]==pixel)
		return outlet;
		outlet=next[outlet];
	}
	return 0;
} // gifcompressor::findpixeloutlet

int message_decoder::length_of_next_phrase_nongreedy()
{	
	unsigned short current_code=NextEntry-1;
	int length=1;
	while (Prefix[current_code] > cc)
	{
		length++;
		current_code=Prefix[current_code];
	}
	return length;
};

void message_decoder::add_phrase_to_trie(unsigned short phrase)
{
	unsigned short down=axon[phrase];
	unsigned char pixel=0;
	if (next_pixel < total_nof_pixels)
	{
		pixel=raster[next_pixel];
	}
	pix[freecode]=pixel;
	axon[freecode]=next[freecode]=0;
	if (!down)
	{
		axon[phrase]=freecode;
	}
	else
	{
		while(next[down]) down=next[down];
		next[down]=freecode;
	}
	freecode++;
	return ;
};

int message_decoder::decode_message()
{
/////debugging information
	//ofstream outputtrie((output_message_file+"mydecoder.trie").c_str(),ios::out|ios::binary);
////debugging information
	while (next_pixel < total_nof_pixels)
	{
		unsigned short Code=next_phrase_nongreedy_from_file();
//////debugging
		//outputtrie << Code << ',';
//////debugging
		nofphrases++;
		if (Code == eoi) break;
		if (Code == cc)
		{
			CodeSize=BPP+1;
			NextEntry=FirstEntry;
			iniroots();
			continue;
		}
		else
		{
			Prefix[NextEntry]=Code;
			NextEntry=NextEntry+1;

			int sizenongreedy=length_of_next_phrase_nongreedy();
			int sizegreedy=length_of_next_phrase_greedy();

			if (NextEntry > (1 << CodeSize))
			{
				CodeSize=CodeSize+1;//need to be checked
			}
				//something to be checked here
			next_pixel=next_pixel+sizenongreedy;
			add_phrase_to_trie(Code);

			if (sizegreedy >= ((1 << bits_length_reduction)+1))
			{
				phrase_counter++;
				//something here;
			
				if (sizenongreedy < sizegreedy)
				{
					nofencodedphrases++;
					//cout <<Code<<','<<NextEntry-1<<endl;
					unsigned int length_reduction=sizegreedy-sizenongreedy;
					length_reduction=length_reduction & ((1<<bits_length_reduction)-1);
					length_reduction=length_reduction << bits_interval;
					phrase_counter=phrase_counter&((1<<bits_interval)-1);
					length_reduction=length_reduction + phrase_counter;
					unsigned long one_message= (*((unsigned long *)(msg + (number_of_bits_of_message / 8))));
					length_reduction=length_reduction << (number_of_bits_of_message % 8);
					one_message=one_message+length_reduction;
					(*((unsigned long *)(msg + (number_of_bits_of_message / 8))))=one_message; //suspicious
					number_of_bits_of_message=number_of_bits_of_message+bits_interval+bits_length_reduction;
					phrase_counter=0;
				}
			}

		}
	}
	cout<<"total number of phrases: "<< nofphrases << endl;
	cout<<"total number of encoding phrases" << nofencodedphrases << endl;
////////debugging
	//outputtrie.close();
////////debugging	
	return 0;
};

int message_decoder::write_message()
{
	ofstream msg_file(output_message_file.c_str(),ios::out|ios::binary);
	msg_file.write((char*)msg,number_of_bits_of_message/8);
	msg_file.close();
	return number_of_bits_of_message;
};

