#include "encoder.h"

unsigned short max(unsigned short a,unsigned short b)
{
	if (a>b)
	{
		return a;
	}
	else 
	{
		return b;
	}
};

unsigned short min(unsigned short a,unsigned short b)
{
	if (a>b)
	{
		return b;
	}
	else 
	{
		return a;
	}
};

void bitpacker::addcodetobuffer(unsigned long code, short n)
{
	unsigned long mask;

	if(n<0)           
	{               
		if(need<8)
		{
			pos++;
			*pos=0x00;              
		}
		need=8;
		return;
	}
	while (n >= need)
	{
		mask = (1<<need)-1;            // 'mask'= all zeroes followed by 'need' ones
		*pos += (unsigned char)((mask&code) << (8-need)); // the 'need' lowest bits of 'code'                                                                                // fill the current byte at its upper end
		pos++;                     // byte is now full, next byte
		*pos=0x00;                 // initialize it
		code = code>>need;         // remove the written bits from code
		n -= need;                 // update the length of 'code'
		need=8;                    // current byte can take 8 bits
	}
	  //                                                        Now we have n < need
	if(n>0)
	{
		mask= (1<<n)-1;
		*pos += (unsigned char)((mask&code)<<(8-need));
			  // (remainder of) code is written to the n
			  // rightmost free bits of the current byte.
		need-=n;                   // The current byte can still take 'need' bits,
			  // and we have 'need'>0. The bits will be filled
			  // upon future calls.
	}                         
	return;
} // bitpacker::addcodetobuffer

void bitpacker::submit(unsigned long code,unsigned short n)
{
	addcodetobuffer(code,n);
	if(pos-buffer >= 255)             // pos pointing to buffer[255] or beyond
	{
		binfile->put((char)0xff);             // write the "bytecount-byte"
		binfile->write((char*)buffer,255);   // write buffer[0..254] to file
		buffer[0]=buffer[255];          // rotate the following bytes,
		buffer[1]=buffer[256];          // which may still contain data, to the
		buffer[2]=buffer[257];          // beginning of buffer, and point
		buffer[3]=buffer[258];          // (pos,need) to the position for new
		pos-=255;                       // input ('need' can stay unchanged)
		byteswritten+=256;
	}
	return;
} // bitpacker::submit

void bitpacker::writeflush(void)
{
	addcodetobuffer(0,-1);        // close any partially filled terminal byte
	if(pos<=buffer)               // buffer is empty
	return;
	binfile->put((char)(pos-buffer));
	binfile->write((char*)buffer,pos-buffer);
	byteswritten+=(pos-buffer+1);
	pos=buffer;
	*pos=0x00;
	need=8;
} // bitpacker::writeflush


bitpacker::bitpacker(ofstream *bf)
{
	binfile=bf;
	need=8;
	buffer=new unsigned char[260];
	byteswritten=0;
	if(buffer)
	{
		pos=buffer;
		*pos=0x00;
		bad=false;
	}
	else
	bad=true;
} // bitpacker-Constructor



bitpacker::~bitpacker(void)
{delete[] buffer;} ;// bitpacker-Destructor

gifcompressor::gifcompressor(LZWoptions option,encoder_application_specification command_line_parameters)
{
	length_of_the_encoding_phrase_greater_than=command_line_parameters.bits_length_reduction;
	every_nof_phrases=command_line_parameters.bits_interval;

	greedyness=option;
	nofphrases=0;
	nofencodedphrases=0;
	nofpixels=0;
	srand(time(0));
	phrase_counter=0;
	randombits=new unsigned char[1024*1024];
	ifstream msgfile(command_line_parameters.input_msg_file.c_str(),ios::in|ios::binary);
	msgfile.read((char*)randombits,1024*1024);
	msgfile.close();

}

gifcompressor::~gifcompressor(void)
{
	delete[] randombits;
}


void gifcompressor::iniroots(void)
{
	unsigned short nofrootcodes=1<<max(2,datadepth);
	for(unsigned short i=0; i<nofrootcodes; i++)
	{
		axon[i]=0;
		pix[i]=(unsigned char)i;
		// next[] is unused for root codes       
	}
} // gifcompressor::iniroots


void gifcompressor::flushstringtable(void)
{
	unsigned short nofrootcodes=1<<max(2,datadepth);
	for(unsigned short i=0; i<nofrootcodes; i++)
		axon[i]=0;
} // gifcompressor::flushstringtable


unsigned short gifcompressor::findpixeloutlet(unsigned short headnode,unsigned char pixel)
{
	unsigned short outlet=axon[headnode];
	while(outlet)
	{
		if (pix[outlet]==pixel)
		return outlet;
		outlet=next[outlet];
	}
	return 0;
} // gifcompressor::findpixeloutlet


unsigned long gifcompressor::weiter(void)
{
	nofphrases++;
	unsigned short up=pixel,down;          // start with the root node for 'pixel'
	if(++curordinal > nofdata)           // end of data stream ? Terminate
	{
		bp->submit(up,nbits);
		return curordinal;
	}                 
	pixel=data[curordinal-1];  
	while((down=findpixeloutlet(up,pixel))!=0)
	{
		up=down;
		if(++curordinal > nofdata)         // end of data stream ? Terminate
		{
			bp->submit(up,nbits);
			return curordinal;
		}
		pixel=data[curordinal-1];
	}
	bp->submit(up,nbits);

	pix[freecode]=pixel;
	axon[freecode]=next[freecode]=0;
	
	down=axon[up];
	if(!down)
		axon[up]=freecode;
	else
	{
		while(next[down])
		down=next[down];
		next[down]=freecode;
	}
	return curordinal;
} // gifcompressor::weiter

unsigned long gifcompressor::weiter_nongreedy()
{
	nofphrases++;

	int unit_message_length=length_of_the_encoding_phrase_greater_than+every_nof_phrases;
	int pos= (nofencodedphrases * unit_message_length) / 8;
	unsigned int random_message = *((unsigned int*)(randombits+pos));

	int start_from_bit=(nofencodedphrases*unit_message_length) % 8;
	
	random_message = random_message >> start_from_bit;

	int mask=(1 << every_nof_phrases)-1;
	
	unsigned int random_interval=random_message & mask;

	if (random_interval == 0)
	{
		random_interval= (1 << every_nof_phrases);
	}
	
	random_message = random_message >> every_nof_phrases;
	
	mask=(1 << length_of_the_encoding_phrase_greater_than)-1;

	unsigned int random_reduce_by=random_message & mask;
	
	if (random_reduce_by == 0)
	{
		random_reduce_by=1<<length_of_the_encoding_phrase_greater_than;
	}

	
	int phrase_length_larger_than=(1<<length_of_the_encoding_phrase_greater_than) +1;

	unsigned short code_length=1;
	unsigned short up=pixel,down;          // start with the root node for 'pixel'
	if(++curordinal > nofdata)           // end of data stream ? Terminate
	{
		bp->submit(up,nbits);
		return curordinal;
	}                 
	pixel=data[curordinal-1];  
	while((down=findpixeloutlet(up,pixel))!=0)
	{
		code_length++;
		up=down;
		if(++curordinal > nofdata)         // end of data stream ? Terminate
		{
			bp->submit(up,nbits);
			return curordinal;
		}
		pixel=data[curordinal-1];
	}

/////////////////////////////
	if (code_length >= phrase_length_larger_than)
	{
		phrase_counter++;
		if (phrase_counter ==random_interval )
		{
			int desired_code_length=code_length-random_reduce_by;
			curordinal=curordinal-code_length;
			up=data[curordinal-1];
			curordinal++;
			pixel=data[curordinal-1];
			code_length=1;
			while(code_length < desired_code_length)
			{
				down=findpixeloutlet(up,pixel);
				code_length++;
				up=down;
				curordinal++;
				pixel=data[curordinal-1];
			}
			nofencodedphrases++;
			phrase_counter=0;
		}
		else
		{
			//submit the phrase to bitpacker as per normal
		}
	}
	else
	{
		//submit the phrase to bit-packer as per normal
	}

/////////////////////////////

	bp->submit(up,nbits);

	pix[freecode]=pixel;
	axon[freecode]=next[freecode]=0;
	
	down=axon[up];
	if(!down)
		axon[up]=freecode;
	else
	{
		while(next[down])
		down=next[down];
		next[down]=freecode;
	}
	return curordinal;
} // gifcompressor::weiter_nongreedy

unsigned long gifcompressor::writedatablocks(ofstream* bf, unsigned char *pixeldata, unsigned long nof, unsigned long wi,unsigned short dd)
{
	nofdata=nof;                // number of pixels in data stream
	width=wi;                   // bitmap width and height in pixels
	height=nofdata/width;
	data=pixeldata;             // the data array
	curordinal=1;               // pixel #1 is next to be processed           
	pixel=data[curordinal-1];      // get pixel #1
	datadepth=dd;               // number of bits per data item (=pixel)
	nbits=max(3,dd+1);          // initial size of compression codes
	cc       = (1<<(nbits-1));  // 'cc' is the lowest code requiring 'nbits' bits
	eoi      = cc+1;            // 'end-of-information'-code
	freecode = (unsigned short)cc+2;    // code of the next entry to be added to the

	bp = new bitpacker(bf);     // object that does the packing of the codes and renders them to the binary file 'bf'
	axon=new unsigned short[4096];
	next=new unsigned short[4096];
	pix=new unsigned char[4096];
	if(!pix || !next || !axon || !bp || bp->bad)
		{delete[] pix; delete[] next; delete[] axon; delete bp; return 0;}
	iniroots();                 // initialize the string table's root nodes

	bp->submit(cc,nbits);       // Submit one 'cc' as the first code

	while(1)
	{
		if (greedyness==Greedy)
		{
			weiter();                 // generates the next code, submits it to 'bp' and
		}
		else if (greedyness==Nongreedy)
		{
			weiter_nongreedy();
		}
		if(curordinal > nofdata)  	// if reached the end of data stream:
		{
			bp->submit(eoi,nbits);  // submit 'eoi' as the last item of the code stream
			bp->writeflush();       // write remaining codes including this 'eoi' to the binary file
			bf->put((char)0x00);    // write an empty data block to signal the end of "raster data" section in the file
			delete[] axon;
			delete[] next;
			delete[] pix;
			unsigned long byteswritten = 2 + bp->byteswritten;
			delete bp;
//debugging purposes
			cout << "after ====================\n";
			cout << "no of phrases:" << nofphrases << "\n";
			cout << "no of encoded phrases:" << nofencodedphrases << "\n";
			cout << "no of bytes written:" << byteswritten << endl;
//for debugging purposes

			return byteswritten;
		}

		if(freecode==(1U<<nbits))     // if the latest code added to the stringtable exceeds 'nbits' bits:
		{
			nbits++;              // increase size of compression codes by 1 bit     
		}
		
		freecode++;

		if(freecode==0xfff)      
		{                    
			flushstringtable();     // avoid stringtable overflow
			bp->submit(cc,nbits);   // tell the decoding software to flush its stringtable     
			nbits=max(3,1+datadepth);     
			freecode=(unsigned short)cc+2;
		}
	} // while(1)

} // gifcompressor::writedatablocks

unsigned long gifcompressor::write_back_message(ofstream* messagefile)
{

	messagefile->write((char*)randombits, (nofencodedphrases*(length_of_the_encoding_phrase_greater_than+every_nof_phrases))/8);
	return nofencodedphrases;
}

unsigned long gifcompressor::get_nofphrases()
{
	return nofphrases;
};

unsigned long gifcompressor::get_nofencodedphrases()
{
	return nofencodedphrases;
};

