// gifencoderdecoder.cpp : Defines the entry point for the console application.

#include <string>
#include <iostream>
#include <fstream>

#include "encoder.h"
#include "somedatastructures.h"
using namespace std;


//*************The set of global variables that represent the status of the application
//*************begin the definition of the global variables*******************
int length_of_the_encoding_phrase_greater_than=2;
int every_nof_phrases=1;
//*************end of the definition of the global variables******************

struct COLOR {unsigned char b,g,r,x;};	// Windows GDI expects 4bytes per color

int LZWDecoder (char * bufIn, char * bufOut,short InitCodeSize, int Width, int Height)
{
	int n;
	int nPixels, maxPixels;			// Output pixel counter

	short CodeSize;					// Current CodeSize (size in bits of codes)
	short ClearCode;				// Clear code : resets decompressor
	short EndCode;					// End code : marks end of information

	long whichBit;					// Index of next bit in bufIn
	long LongCode;					// Temp. var. from which Code is retrieved
	short Code;						// Code extracted
	short PrevCode;					// Previous Code
	short OutCode;					// Code to output

	// Translation Table:
	short Prefix[0x2000];				// Prefix: index of another Code
	unsigned char Suffix[0x2000];		// Suffix: terminating character
	short FirstEntry;				// Index of first free entry in table
	short NextEntry;				// Index of next free entry in table

	unsigned char OutStack[0x2000];	// Output buffer
	int OutIndex;					// Characters in OutStack



	// Set up values that depend on InitCodeSize Parameter.
	CodeSize = InitCodeSize+1;
	ClearCode = (1 << InitCodeSize);
	EndCode = ClearCode + 1;
	NextEntry = FirstEntry = ClearCode + 2;

	whichBit=0;
	nPixels = 0;
	maxPixels = Width*Height;

	while (nPixels<maxPixels) 
	{
		OutIndex = 0;							// Reset Output Stack
		// GET NEXT CODE FROM bufIn:
		// LZW compression uses code items longer than a single byte.
		// For GIF Files, code sizes are variable between 9 and 12 bits 
		// That's why we must read data (Code) this way:
		LongCode=*((long*)(bufIn+whichBit/8));	// Get some bytes from bufIn
		LongCode>>=(whichBit&7);				// Discard too low bits
		Code =(LongCode & ((1<<CodeSize)-1) );	// Discard too high bits
		whichBit += CodeSize;					// Increase Bit Offset

		// SWITCH, DIFFERENT POSIBILITIES FOR CODE:
		if (Code == EndCode)					// END CODE
			break;								// Exit LZW Decompression loop

		if (Code == ClearCode) {				// CLEAR CODE:
			CodeSize = InitCodeSize+1;			// Reset CodeSize
			NextEntry = FirstEntry;				// Reset Translation Table
			PrevCode=Code;				// Prevent next to be added to table.
			continue;							// restart, to get another code
		}
		if (Code < NextEntry)
		{	
			OutCode = Code;						// Set code to output.
		}
		else 
		{									// CODE IS NOT IN TABLE:
			OutIndex++;			// Keep "first" character of previous output.
			OutCode = PrevCode;					// Set PrevCode to be output
		}

		// EXPAND OutCode IN OutStack
		// - Elements up to FirstEntry are Raw-Codes and are not expanded
		// - Table Prefices contain indexes to other codes
		// - Table Suffices contain the raw codes to be output
		while (OutCode >= FirstEntry) 
		{
			if (OutIndex > 4096) 
			{
				return 0;
			}
			OutStack[OutIndex++] = Suffix[OutCode];	// Add suffix to Output Stack
			OutCode = Prefix[OutCode];				// Loop with preffix
		}

		// NOW OutCode IS A RAW CODE, ADD IT TO OUTPUT STACK.
		if (OutIndex > 4096)
		{
			return 0;
		}

		OutStack[OutIndex++] = (unsigned char) OutCode;

		if (Code==NextEntry)
		{
			OutStack[0]=OutCode;
		}

		// ADD NEW ENTRY TO TABLE (PrevCode + OutCode)
		// (EXCEPT IF PREVIOUS CODE WAS A CLEARCODE)
		if (PrevCode!=ClearCode) 
		{
			Prefix[NextEntry] = PrevCode;
			Suffix[NextEntry] = (unsigned char) OutCode;
			NextEntry++;

     
			// INCREASE CodeSize IF NextEntry IS INVALID WITH CURRENT CodeSize
			if (NextEntry >= (1<<CodeSize)) 
			{
				if (CodeSize < 12) 
				{
					CodeSize++;
				}

			}
		}

		PrevCode = Code;

		// Avoid the possibility of overflow on 'bufOut'.
		if (nPixels + OutIndex > maxPixels) OutIndex = maxPixels-nPixels;

		// OUTPUT OutStack (LAST-IN FIRST-OUT ORDER)
		for (n=OutIndex-1; n>=0; n--) 
		{

			bufOut[nPixels]=OutStack[n];	// Write output
			nPixels++;					// Increase counters.
		}

	}	// while (main decompressor loop)

	return whichBit;
}


int LoadGIF (encoder_application_specification command_line_parameters,LZWoptions option)
{
	string szFileName=command_line_parameters.input_gif_file;
	string szOutFileName=command_line_parameters.output_gif_file;
	int n;
	char junk[4096];
	gifcompressor relaxedLZWcompressor(option,command_line_parameters);
	// Global GIF variables:
	int GlobalBPP;						// Bits per Pixel.
	COLOR * GlobalColorMap;				// Global colormap (allocate)

	struct GIFGCEtag {				// GRAPHIC CONTROL EXTENSION
		unsigned char BlockSize;		// Block Size: 4 bytes
		unsigned char PackedFields;		// Packed Fields. Bits detail:
										//    0: Transparent Color Flag
										//    1: User Input Flag
										//  2-4: Disposal Method
		unsigned short Delay;			// Delay Time (1/100 seconds)
		unsigned char Transparent;		// Transparent Color Index
	}__attribute__((__packed__)) gifgce;

	int current_pos=0; //for debugging purposes
	int current_pos1= 0; 
	// OPEN FILE
	ifstream giffile (szFileName.c_str(),ios::in|ios::binary);
	ofstream giffileout(szOutFileName.c_str(),ios::out|ios::binary);
	
	if (!giffile.is_open()) 
	{
		return 0;
	}


	// *1* READ HEADER (SIGNATURE + VERSION)
	char szSignature[6];				// First 6 bytes (GIF87a or GIF89a)
	giffile.read(szSignature,6);


	if ( memcmp(szSignature,"GIF",2) != 0)
	{ 
		return 0; 
	}


	// *2* READ LOGICAL SCREEN DESCRIPTOR
	struct GIFLSDtag {
		unsigned short ScreenWidth;		// Logical Screen Width
		unsigned short ScreenHeight;	// Logical Screen Height
		unsigned char PackedFields;		// Packed Fields. Bits detail:
										//  0-2: Size of Global Color Table
										//    3: Sort Flag
										//  4-6: Color Resolution
										//    7: Global Color Table Flag
		unsigned char Background;		// Background Color Index
		unsigned char PixelAspectRatio;	// Pixel Aspect Ratio
	}__attribute__((__packed__)) giflsd;


	giffile.read((char*)&giflsd,sizeof(giflsd));
	
	cout<< "Screen Width:"<< giflsd.ScreenWidth<<endl;
	cout<< "Screen Height:"<< giflsd.ScreenHeight<<endl;
	
	GlobalBPP = (giflsd.PackedFields & 0x07) + 1;
	
	cout<< "bits per pixel:"<<GlobalBPP<<endl;

	// fill some animation data:
	int FrameWidth = giflsd.ScreenWidth;
	int FrameHeight = giflsd.ScreenHeight;
	
	// *3* READ/GENERATE GLOBAL COLOR MAP
	GlobalColorMap = new COLOR [1<<GlobalBPP];

	if (giflsd.PackedFields & 0x80)	// File has global color map?
		for (n=0;n< 1<<GlobalBPP;n++)
		{
			GlobalColorMap[n].r=giffile.get();
			GlobalColorMap[n].g=giffile.get();
			GlobalColorMap[n].b=giffile.get();
		}
	

	current_pos = giffile.tellg();
	giffile.seekg(0);
	giffile.read(junk,current_pos);
	giffileout.write(junk,current_pos);
	
	int no_of_frames=0;

	// *4* NOW WE HAVE 3 POSSIBILITIES:
	//  4a) Get and Extension Block (Blocks with additional information)
	//  4b) Get an Image Separator (Introductor to an image)
	//  4c) Get the trailer Char (End of GIF File)
	do
	{
		int charGot = giffile.get();

		if (charGot == 0x21)		// *A* EXTENSION BLOCK 
		{
			switch (giffile.get())
			{

			case 0xF9:			// Graphic Control Extension

				giffile.read((char*)&gifgce,sizeof(gifgce));
				giffile.get(); // Block Terminator (always 0)
				break;

			case 0xFE:			// Comment Extension: Ignored
			case 0x01:			// PlainText Extension: Ignored
			case 0xFF:			// Application Extension: Ignored
			default:			// Unknown Extension: Ignored
				// read (and ignore) data sub-blocks
				while (int nBlockLength = giffile.get())
					for (n=0;n<nBlockLength;n++) giffile.get();
				break;
			}
			current_pos1 = giffile.tellg();
			giffile.seekg(current_pos);
			giffile.read(junk,current_pos1 - current_pos);
			giffileout.write(junk,current_pos1 - current_pos);
			current_pos = current_pos1;

		}


		else if (charGot == 0x2c) // *B* IMAGE (0x2c Image Separator)
		{	
			
			no_of_frames++;

			// Read Image Descriptor
			struct GIFIDtag {	
				unsigned short xPos;			// Image Left Position
				unsigned short yPos;			// Image Top Position
				unsigned short Width;			// Image Width
				unsigned short Height;			// Image Height
				unsigned char PackedFields;		// Packed Fields. Bits detail:
											//  0-2: Size of Local Color Table
											//  3-4: (Reserved)
											//    5: Sort Flag
											//    6: Interlace Flag
											//    7: Local Color Table Flag
			}__attribute__((__packed__)) gifid;

			giffile.read((char*)&gifid, sizeof(gifid));

			int LocalColorMap = (gifid.PackedFields & 0x80)? 1 : 0;

			int BytesPerRow = gifid.Width;

			if (GlobalBPP==24)
				{BytesPerRow*=3;}
			
			
			unsigned char* raster = new unsigned char [BytesPerRow*gifid.Height];
		
			if (LocalColorMap)		// Read Color Map (if descriptor says so)
				giffile.read( (char*)junk, 3*(1<<GlobalBPP));

			short firstbyte = giffile.get();	// 1st byte of img block (CodeSize)

			current_pos1 = giffile.tellg();
			giffile.seekg(current_pos);
			giffile.read(junk,current_pos1 - current_pos);
			giffileout.write(junk,current_pos1 - current_pos);
			current_pos = current_pos1;

			// Calculate compressed image block size
				// to fix: this allocates an extra byte per block
			long ImgStart,ImgEnd;				

			ImgEnd = ImgStart = giffile.tellg();
			while (n=giffile.get()) giffile.seekg (ImgEnd+=n+1);
			giffile.seekg (ImgStart);

			// Allocate Space for Compressed Image
			char * pCompressedImage = new char [ImgEnd-ImgStart+4];
  
			// Read and store Compressed Image
			char * pTemp = pCompressedImage;
			while (int nBlockLength = giffile.get())
			{
				giffile.read(pTemp,nBlockLength);
				pTemp+=nBlockLength;
			}

			// Call LZW/GIF decompressor
			n=LZWDecoder(
				(char*) pCompressedImage,
				(char*) raster,
				firstbyte, 
				gifid.Width, gifid.Height);
			// Some cleanup
			relaxedLZWcompressor.writedatablocks(&(giffileout),raster,gifid.Width*gifid.Height,gifid.Width,firstbyte);

			delete[] pCompressedImage;
			delete[] raster;
			current_pos = giffile.tellg();
		}

		else if (charGot == 0x3b) 
		{	// *C* TRAILER: End of GIF Info
			giffileout.put((unsigned char)0x3b);
			break; // Ok. Standard End.
		}

	} while (giffile.good());

	giffile.close();
	giffileout.close();

	cout<< "number of frames:"<<no_of_frames<<endl;
	if (option==Nongreedy)
	{
		ofstream messagefile(command_line_parameters.msg_taken_in.c_str(),ios::out|ios::binary);
		relaxedLZWcompressor.write_back_message(&messagefile);
		cout << "total number of phrases nongreedy:"<<relaxedLZWcompressor.get_nofphrases()<<"\n";
		cout << "total number of encoded phrases:" << relaxedLZWcompressor.get_nofencodedphrases() << "\n";
		messagefile.close();
	}
	else
	{
		cout << "total number of phrases greedy:"<<relaxedLZWcompressor.get_nofphrases()<<"\n";
	}

	return 0;
}

int main(int argc, char* argv[])
{
	
	string filepath1=argv[1];
	string filepath3=argv[2];
	string filepath2=filepath3.substr(0,filepath3.length()-4)+".greedy.gif";
	string messagefile=argv[3];
//the options are taken as global variable
	length_of_the_encoding_phrase_greater_than=atoi(argv[4]);
	every_nof_phrases=atoi(argv[5]);
	
	string msg_taken_in=argv[6];
//	string information_file=argv[7];
	

	encoder_application_specification command_line_parameters;
	command_line_parameters.input_gif_file=filepath1;
	command_line_parameters.output_gif_file=filepath2;
	command_line_parameters.input_msg_file=messagefile;
	command_line_parameters.bits_interval=every_nof_phrases;
	command_line_parameters.bits_length_reduction=length_of_the_encoding_phrase_greater_than;
	command_line_parameters.msg_taken_in=msg_taken_in;
//	command_line_parameters.information_file=information_file;

/*
	string filepath1="giffile/airplane";
	filepath1=filepath1 + ".gif";
	string filepath2="giffile/airplane";
	filepath2=filepath2 + ".greedy.gif";
	string filepath3="giffile/airplane";
	filepath3=filepath3 + ".nongreedy.gif";

//the options are taken as global variable
	length_of_the_encoding_phrase_greater_than=1;
	every_nof_phrases=4;
*/

	LoadGIF (command_line_parameters,Greedy);

	command_line_parameters.output_gif_file=filepath3;

	LoadGIF (command_line_parameters,Nongreedy);
	return 0;
}
